/* myproject1 project which demonstrates the basics on how to setup a project 
 * example application.
 * Host code.
 */

// includes, system
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

// includes, project
#include <cutil.h>

// includes, kernels
#include <myproject1_kernel.cu>

////////////////////////////////////////////////////////////////////////////////
// declaration, forward
void runTest( int argc, char** argv);

extern "C"
void computeGold( float* reference, float* idata, const unsigned int len);

////////////////////////////////////////////////////////////////////////////////
// Program main
////////////////////////////////////////////////////////////////////////////////
int
main( int argc, char** argv) 
{
    runTest( argc, argv);

    CUT_EXIT(argc, argv);
}

////////////////////////////////////////////////////////////////////////////////
//! Run a simple test for CUDA
////////////////////////////////////////////////////////////////////////////////
void
runTest( int argc, char** argv) 
{

  CUT_DEVICE_INIT(argc, argv);

  unsigned int timer = 0;
  unsigned int timer_host = 0;
  CUT_SAFE_CALL( cutCreateTimer( &timer));
  CUT_SAFE_CALL( cutCreateTimer( &timer_host));

  unsigned int threads_per_block=768/2;  // to schedule 2 blocks per SM

  unsigned int elements = 1000;
  while (elements < 1024*1024)  // outer loop to increment array size
  {
    printf( "\nNumber of elements sorted: %d\n", elements );

    // compute number of elements per block (rounding up to block size)
    // and include extra padding for an overlap by one half block
    unsigned int elements_per_block = threads_per_block*2;
    unsigned int elements_padded = (elements+elements_per_block-1)
                                   /(elements_per_block)
                                   *(elements_per_block);
    unsigned int elements_per_halfblock = elements_per_block/2;
    unsigned int elements_total = elements_padded;

    unsigned int mem_size = sizeof(float) * elements_total;

    // allocate host memory
    float* h_idata = (float*) malloc( mem_size);

    // initalize the memory
    unsigned int i;
    for( i = 0; i < elements; ++i)   // generate random data to sort
        h_idata[i] = (float) rand() / RAND_MAX;

    for(      ; i < elements_total; ++i)  // padding with largest IEEE754 float number
        h_idata[i] = 3.402823466e+38F;

    // allocate device memory
    float* d_data;
    CUDA_SAFE_CALL( cudaMalloc( (void**) &d_data, mem_size));

    // allocate mem for the result on host side
    float* h_odata = (float*) malloc( mem_size);

    // setup execution parameters
    unsigned int num_blocks = ((elements_padded+1)/2+threads_per_block-1)/threads_per_block;
    dim3  grid( num_blocks, 1, 1);
    dim3  grid2( num_blocks-1, 1, 1);
    dim3  threads( threads_per_block, 1, 1);

    printf( "Grid is %d blocks wide, block has %d threads\n", num_blocks, threads_per_block );

    // outer loop
    unsigned int max_iterations = num_blocks;

    CUT_SAFE_CALL( cutStartTimer( timer));

    // copy host memory to device
    CUDA_SAFE_CALL( cudaMemcpy( d_data, h_idata, mem_size,
                                cudaMemcpyHostToDevice) );

    // we perform the amount of iterations that guarantee that even in worst
    // case an element can traverse from top to bottom and vice versa.
    for (i=0; i<max_iterations; i++)
    {
        // execute the kernel working in place on d_data
        testKernel<<< grid, threads, elements_per_block*sizeof(float) >>>
                             ( d_data );

        // check if kernel execution generated an error
        CUT_CHECK_ERROR("Kernel execution failed (even iteration)");

        if (grid2.x > 0)
        {
            // execute the kernel working in place on d_data
            // offset by a half block in terms of elements
            testKernel<<< grid2, threads, elements_per_block*sizeof(float) >>>
                                 (&d_data[elements_per_halfblock]);

            // check if kernel execution generated an error
            CUT_CHECK_ERROR("Kernel 2 execution failed (odd iteration)");
        }
    }

    // copy result from device to host
    CUDA_SAFE_CALL( cudaMemcpy( h_odata, d_data, mem_size,
                                cudaMemcpyDeviceToHost) );

    CUT_SAFE_CALL( cutStopTimer( timer));

    printf( "Processing time of odd-even sort on GPU: %f (ms)\n", cutGetTimerValue( timer));
    
    // compute reference solution
    float* reference = (float*) malloc( mem_size);

    CUT_SAFE_CALL( cutStartTimer( timer_host));

    computeGold( reference, h_idata, elements);

    CUT_SAFE_CALL( cutStopTimer( timer_host));

    printf( "Processing time of quicksort on host CPU: %f (ms)\n", cutGetTimerValue( timer_host));

    printf( "GPU vs. CPU speedup factor: %f\n", cutGetTimerValue( timer_host) / cutGetTimerValue( timer));

    // check if the result is equivalent to the expected soluion
    CUTBoolean res = cutComparef( reference, h_odata, elements);
    printf( "Test %s\n", (1 == res) ? "PASSED" : "FAILED");
    
    // cleanup memory
    free( h_idata);
    free( h_odata);
    free( reference);
    CUDA_SAFE_CALL(cudaFree(d_data));

    if (!res) break;

    elements *= 10;
  }

  CUT_SAFE_CALL( cutDeleteTimer( timer_host));
  CUT_SAFE_CALL( cutDeleteTimer( timer));
}
