/* myproject1 project which demonstrates the basics on how to setup a project 
 * example application.
 * Device code.
 */

#ifndef _MYPROJECT1_KERNEL_H_
#define _MYPROJECT1_KERNEL_H_

#include <stdio.h>

////////////////////////////////////////////////////////////////////////////////
//! Simple test kernel for device functionality
//! @param g_idata  input data in global memory
//! @param g_odata  output data in global memory
////////////////////////////////////////////////////////////////////////////////

#define SDATA( index)      CUT_BANK_CHECKER(sdata, index)

// This swap first gets the argument from shared memory a,b
// into registers A,B, compares the registers and conditionally
// writes the registers back to shared memory in swapped order.
// A swap flag is also set conditionally.
#define cond_swap(a,b,c) { float A = (a), B = (b); if (A > B) { (a) = B ; (b) = A; (c)=true; } }


__global__ void testKernel(float *inout)
{
__shared__ int swapped[8];  // 8 integer flags for bank-conflict-free access

// shared memory
// the size in bytes is determined by the host application as
// the third argument in kernel invocation
extern  __shared__  float sdata[];

// compute device memory source/destination data pointers
// the 24 bit multiplication is faster, but limits us to 16M elements
float *pio1 = &inout[__umul24(blockDim.x,blockIdx.x)*2 + threadIdx.x];
float *pio2 = pio1 + blockDim.x;

// compute shared data array pointers for data copy operation
float *ps1 = &sdata[threadIdx.x/2 + ((threadIdx.x & 1) ? blockDim.x : 0)];
float *ps2 = ps1 + blockDim.x/2;

// copy input data to shared memory (coalesced read), 2 floats per thread
// I separate odd and even elements to avoid bank conflicts later on.
// the write to SDATA causes 2-way bank conflicts however!
*ps1 = *pio1;
*ps2 = *pio2;

// compute per-thread pointers for our per-block shared data array
float *plidx_even  = &sdata[threadIdx.x];
float *plidx_even1 = plidx_even + 1;
float *plidx_odd   = plidx_even + blockDim.x;

// a trick to disable odd-even swapping on last thread because it would
// spill into the neighbor block (no cross block-synchronization!)
if(threadIdx.x == blockDim.x-1) plidx_even1 = plidx_odd;

// each thread in a half-warp will write to a swap flag in a different memory bank
int *pswapped      = &swapped[threadIdx.x & 7];

// This computation now avoids bank conflicts, however the large
// number of syncthreads make it rather inefficient.
// I unrolled the loop a bit to gain more speed. An exit condition
// check is only performed once per 8 full swaps, after the first
// full swap in the while loop. A full swap consists of an even-odd
// half swap and one odd-even half swap.
while(true)
{
  __syncthreads();  
  // the 8 master threads set swap flag for each bank to false 
  if (threadIdx.x < 8) *pswapped = false;

  // swap even with odd elements and odd with even+1 elements
  __syncthreads(); cond_swap(*plidx_even, *plidx_odd,   *pswapped);
  __syncthreads(); cond_swap(*plidx_odd,  *plidx_even1, *pswapped);

  __syncthreads();
  // check for exit condition from all 8 memory banks after the first swap round
  if (!(swapped[0]||swapped[1]||swapped[2]||swapped[3]||swapped[4]||swapped[5]||swapped[6]||swapped[7])) break;

  // swap even with odd elements and odd with even+1 elements 7 more times
  __syncthreads(); cond_swap(*plidx_even, *plidx_odd,   *pswapped);
  __syncthreads(); cond_swap(*plidx_odd,  *plidx_even1, *pswapped);
  __syncthreads(); cond_swap(*plidx_even, *plidx_odd,   *pswapped);
  __syncthreads(); cond_swap(*plidx_odd,  *plidx_even1, *pswapped);
  __syncthreads(); cond_swap(*plidx_even, *plidx_odd,   *pswapped);
  __syncthreads(); cond_swap(*plidx_odd,  *plidx_even1, *pswapped);
  __syncthreads(); cond_swap(*plidx_even, *plidx_odd,   *pswapped);
  __syncthreads(); cond_swap(*plidx_odd,  *plidx_even1, *pswapped);
  __syncthreads(); cond_swap(*plidx_even, *plidx_odd,   *pswapped);
  __syncthreads(); cond_swap(*plidx_odd,  *plidx_even1, *pswapped);
  __syncthreads(); cond_swap(*plidx_even, *plidx_odd,   *pswapped);
  __syncthreads(); cond_swap(*plidx_odd,  *plidx_even1, *pswapped);
  __syncthreads(); cond_swap(*plidx_even, *plidx_odd,   *pswapped);
  __syncthreads(); cond_swap(*plidx_odd,  *plidx_even1, *pswapped);
}

__syncthreads();

// copy data from shared memory to output (coalesced write)
// merging odd and even elements again
// the read from SDATA causes 2-way bank conflicts!
*pio1 = *ps1;
*pio2 = *ps2;
}

#endif // #ifndef _MYPROJECT1_KERNEL_H_
